# Helpers for implementing anvil.server on an (optionally threaded) Real Python process.
# Used in uplink and downlink, and now even in the PyPy sandbox.

import random, string, json, re, sys, time, importlib, anvil


# For single-threaded implementations, re-entrant calls occupy the same thread,
# so we fake "switching a thread" by pushing all the thread-locals onto a stack
# while we're handling an inner call
class StackableLocal(object):
    def _push_stack(self):
        _nested = dict(self.__dict__)
        self.__dict__.clear()
        self.__dict__['_nested'] = _nested
        self.__init__()

    def _pop_stack(self):
        nested = self._nested
        self.__dict__.clear()
        self.__dict__.update(nested)

    def __new__(cls, *args, **kwargs):
        v = object.__new__(cls)
        cls.__init__(v, *args, **kwargs)
        _stackables.append(v)
        return v


_stackables = []

try:
    import threading
    ThreadLocal = threading.local
    MULTITHREADED = True
except:
    MULTITHREADED = False
    ThreadLocal = StackableLocal


from . import  _serialise, _server
from ._server import LazyMedia, registrations

string_type = str if sys.version_info >= (3,) else basestring

console_output = sys.stdout

class HttpRequest(ThreadLocal):

    def __init__(self):
        self._prevent_access = True

    def __getattribute__(self, name):
        if name not in ['_push_stack', '_pop_stack', '_nested', '__dict__', '__init__'] and \
                ThreadLocal.__getattribute__(self, "_prevent_access"):
            raise Exception("anvil.server.request is only available in http_endpoint calls.")

        return ThreadLocal.__getattribute__(self, name)

    @property
    def body_json(self):
        if hasattr(self, "_body_json"):
            return self._body_json
        elif self.body is not None and self.headers.get("content-type", None) == "application/json":
            self._body_json = json.loads(self.body.get_bytes())
        else:
            self._body_json = None
        return self._body_json


_server.api_request = HttpRequest()


def gen_id():
    sr = random.SystemRandom()
    chars = string.ascii_letters + string.digits
    return ''.join(sr.choice(chars) for _ in range(10))


# Overwrite with functions from context
send_reqresp = None


class LocalCallInfo(ThreadLocal):
    def __init__(self):
        self.call_id = None
        self.stack_id = None
        self.session = None
        self.cache_filter = {}
        self.cache_update = {}
        self.dump_task_state = False
        self.enable_profiling = False

    def __getitem__(self, item):
        return self.session.__getitem__(item)

    def __setitem__(self, key, value):
        return self.session.__setitem__(key, value)

    def __delitem__(self, key):
        del self.session[key]

    def get(self, key, default=None):
        return self.session.get(key, default)

    def __iter__(self):
        return self.session.__iter__()

    def __repr__(self):
        return "<Session:%s>" % repr(self.session)


class LocalCallContext(_server.CallContext, ThreadLocal):
    def __init__(self):
        self._setup(None, [])


call_info = LocalCallInfo()
call_context = LocalCallContext()
call_responses = {}
waiting_for_calls = threading.Condition() if MULTITHREADED else None


# If MULTITHREADED is False, better overwrite this
def poll_for_call_responses(*args):
    raise AssertionError("We're in single-threaded mode, but poll_for_call_responses() is not set")


backends = {}


def _switch_session():
    import anvil.server
    sjson = anvil.server.call('anvil.private.switch_session!') or {"session": {}, "objects": []}
    call_info.session = _server._reconstruct_objects(sjson, None, remote_is_trusted=True)["session"]


default_app = anvil.app


class LocalAppInfo(ThreadLocal):
    def __getattr__(self, attr):
        return getattr(default_app, attr)

    def __setattr__(self, key, value):
        raise AttributeError("This object is read-only")

    def _setup(self, environment={}, **kwargs):
        self.__dict__.update(kwargs, environment=anvil._AppInfo._Environment(**environment))


anvil.app = LocalAppInfo()


class SendNoResponse(Exception):
    pass


class IncomingRequest(_serialise.IncomingReqResp):
    def __init__(self, json, import_modules=None, run_fn=None, dump_task_state=False, setup_task_state=None):
        self.import_modules = import_modules
        self.run_fn = run_fn
        self.dump_task_state = dump_task_state
        self.setup_task_state = setup_task_state
        remote_is_trusted = _server.CallContext.StackFrame(json.get('call-stack', [{}])[0]).is_trusted
        _serialise.IncomingReqResp.__init__(self, json, remote_is_trusted=remote_is_trusted)

    def execute(self):
        def make_call():
            call_info.call_id = self.json.get('id')
            call_info.stack_id = self.json.get('call-stack-id', None)
            call_info.session_id = self.json.get('session-id')
            sjson = self.json.get('sessionData', {'session': None, 'objects': []})
            call_info.session = None
            call_info.enable_profiling = self.json.get('enable-profiling', False)
            if call_info.enable_profiling:
                call_info.profile = {
                    "origin": "Server (Python)",
                    "description": "Python _threaded_server execution",
                    "start-time": time.time()*1000,
                }
            call_info.cache_filter = _server.get_liveobject_cache_filter_spec([self.json.get('args'), self.json.get('kwargs')])
            call_info.cache_update = {}
            call_info.dump_task_state = self.dump_task_state
            call_context._setup(self.json.get('client', {}), self.json.get('call-stack'))
            anvil.app._setup(**self.json.get('app-info', {}))
            if self.setup_task_state:
                self.setup_task_state(call_info.call_id, True)
            import_complete = False
            try:
                if self.import_modules:
                    self.import_modules()
                import_complete = True

                # Now we've imported enough to deserialise custom types
                self.reconstruct_remaining_data()
                call_info.session = _server._reconstruct_objects(sjson, None, remote_is_trusted=self.remote_is_trusted).get("session", {})

                if self.run_fn is not None:
                    response = self.run_fn()
                elif 'liveObjectCall' in self.json:
                    loc = self.json['liveObjectCall']
                    spec = dict(loc)

                    if call_context.remote_caller is None:
                        spec["source"] = "UNKNOWN"
                    elif call_context.remote_caller.is_trusted:
                        spec["source"] = "server"
                    else:
                        spec["source"] = "client"

                    del spec["method"]
                    backend = loc['backend']
                    if backend not in backends:
                        raise Exception("No such LiveObject backend: " + repr(backend))
                    inst = backends[backend](spec)
                    method = getattr(inst, loc['method'])

                    call_info.cache_filter.setdefault(backend, set()).add(spec['id'])

                    response = method(*self.json['args'], **self.json['kwargs'])
                else:
                    command = self.json['command']
                    for reg in registrations:
                        m = re.match(reg, command)
                        if m and len(m.group(0)) == len(command):
                            response = registrations[reg](*self.json["args"], **self.json["kwargs"])
                            break
                    else:
                        if self.json.get('stale-uplink?'):
                            raise _server.UplinkDisconnectedError({'type': 'anvil.server.UplinkDisconnectedError',
                                                                   'message':'The uplink server for "%s" has been disconnected' % command})

                        else:
                            raise _server.NoServerFunctionError({'type': 'anvil.server.NoServerFunctionError',
                                                                 'message': 'No server function matching "%s" has been registered' % command})

                def err(*args):
                    raise Exception("Cannot save DataMedia objects in anvil.server.session")

                try:
                    sjson = _server.fill_out_media({'session': call_info.session}, err, remote_is_trusted=True)
                    json.dumps(sjson)
                except TypeError as e:
                    raise _server.SerializationError("Tried to store illegal value in a anvil.server.session. " + e.args[0])
                except _server.SerializationError as e:
                    raise _server.SerializationError("Tried to store illegal value in a anvil.server.session. " + e.args[0])

                resp = {"id": self.json["id"], "response": response, "sessionData": sjson, "cacheUpdates": call_info.cache_update}

                if call_info.enable_profiling:
                    call_info.profile["end-time"] = time.time()*1000
                    resp["profile"] = call_info.profile

                _server.fill_out_cap_updates(resp, self.capabilities)

                if self.dump_task_state:
                    try:
                        task_state = dict(anvil.server.task_state)
                        tjson = _server.fill_out_media({'taskState': task_state}, err, remote_is_trusted=True)
                        json.dumps(tjson)
                        resp['taskState'] = task_state
                    except (TypeError, _server.SerializationError):
                        pass

                try:
                    send_reqresp(resp, remote_is_trusted=call_context.remote_caller.is_trusted)
                except _server.SerializationError as e:
                    raise _server.SerializationError("Cannot serialize return value from function. " + str(e))
            except SendNoResponse:
                pass
            except:

                e = _server._report_exception(self.json["id"])

                if self.dump_task_state:
                    def err(*args):
                        raise Exception("Cannot save DataMedia objects in anvil.server.session")

                    try:
                        task_state = dict(anvil.server.task_state)
                        tjson = _server.fill_out_media({'taskState': task_state}, err, remote_is_trusted=True)
                        json.dumps(tjson)
                    except (TypeError, _server.SerializationError):
                        pass
                    else:
                        e['taskState'] = task_state

                if not import_complete:
                    e['moduleLoadFailed'] = True

                try:
                    send_reqresp(e)
                except:
                    trace = "\ncalled from ".join(["%s:%s" % (t[0], t[1]) for t in e["error"]["trace"]])
                    console_output.write(("Failed to report exception: %s: %s\nat %s\n" % (e["error"]["type"], e["error"]["message"], trace)).encode("utf-8"))
                    console_output.flush()
            finally:
                if self.setup_task_state:
                    self.setup_task_state(call_info.call_id, False)
                self.complete()

        if MULTITHREADED:
            threading.Thread(target=make_call).start()
        else:
            make_call()

    def complete(self):
        pass


class IncomingResponse(_serialise.IncomingReqResp):
    def __init__(self, json):
        # Responses from the server are trusted by definition
        _serialise.IncomingReqResp.__init__(self, json, remote_is_trusted=True)

    def execute(self):
        id = self.json['id']
        if id in call_responses:
            call_responses[id] = (self, self.json)
            if MULTITHREADED:
                with waiting_for_calls:
                    waiting_for_calls.notifyAll()
        else:
            print("Got a response for an unknown ID: " + repr(self.json))


def kill_outstanding_requests(msg):
    for k in call_responses.keys():
        if call_responses[k] is None:
            call_responses[k] = (None, {'error': {'message': msg}})

    if not MULTITHREADED:
        raise Exception("_threaded_server.kill_outstanding_requests() does not work in single-threaded mode")

    with waiting_for_calls:
        waiting_for_calls.notifyAll()


def register_live_object_backend(cls):

    name = "uplink." + cls.__name__
    backends[name] = cls

    if _server.on_register is not None:
        _server.on_register(name, True)

    return cls


live_object_backend = register_live_object_backend


def do_call(args, kwargs, fn_name=None, live_object=None): # Yes, I do mean args and kwargs without *s
    id = gen_id()

    call_responses[id] = None

    capabilities_for_update = []

    if call_info.enable_profiling:
        profile = {
            "origin": "Server (Python)",
            "description": "Outgoing call from Python _threaded_server",
            "start-time": time.time()*1000
        }

    def send_call():
        # print("Call stack ID = " + repr(_call_info.stack_id))
        if call_info.stack_id is None:
            call_info.stack_id = "outbound-" + gen_id()
        req = {'type': 'CALL', 'id': id, 'args': args, 'kwargs': kwargs,
               'call-stack-id': call_info.stack_id, 'originating-call': call_info.call_id}

        if live_object:
            req["liveObjectCall"] = { k: live_object._spec[k] for k in ["id", "backend", "mac", "permissions"] }
            req["liveObjectCall"]["method"] = fn_name
        elif fn_name:
            req["command"] = fn_name
        else:
            raise Exception("Expected one of fn_name or live_object")
        try:
            # We're calling a server function and those are always trusted
            send_reqresp(req, collect_capabilities=capabilities_for_update, remote_is_trusted=True)
        except _server.SerializationError as e:
            raise _server.SerializationError("Cannot serialize arguments to function. " + str(e))

    if MULTITHREADED:
        with waiting_for_calls:
            send_call()
            while call_responses[id] is None:
                waiting_for_calls.wait()
    else:
        send_call()
        dump_task_state = call_info.dump_task_state
        # Fake a thread switch
        for s in _stackables:
            s._push_stack()
        while call_responses[id] is None:
            poll_for_call_responses(dump_task_state)
            dump_task_state = False # only do it first time
        for s in _stackables:
            s._pop_stack()

    if call_info.enable_profiling:
        profile["end-time"] = time.time()*1000

    reqresp, r = call_responses.pop(id)

    # Now we're in the right thread, we can do any custom deserialisation
    if reqresp:
        reqresp.reconstruct_remaining_data()

    if "cacheUpdates" in r:
        # Apply updates to any of our own objects that were passed in
        _server.apply_cache_updates(r['cacheUpdates'], [args, kwargs, live_object])
        # Queue up whichever updates *we* should be returning
        _server.combine_cache_updates(call_info.cache_update, r['cacheUpdates'], call_info.cache_filter)

    _server.apply_cap_updates(r, capabilities_for_update)

    if call_info.enable_profiling:
        if "profile" in r:
            profile["children"] = [r["profile"]]

        if hasattr(call_info, "profile"):
            if "children" not in call_info.profile:
                call_info.profile["children"] = []

            call_info.profile["children"].append(profile)

    if 'response' in r:
        return r['response']
    if 'error' in r:
        error_from_server = _server._deserialise_exception(r["error"])
        raise error_from_server
    else:
        raise Exception("Bogus response from server: " + repr(r))



from anvil.server import portable_class

from ._helpers import _hash_wrapper

# Don't load v2 code unless v2 is imported. v2._load_hacks will inject this for us.
# from .v2._refs import make_refs as _make_refs
_make_refs = lambda x: x


# Hack: Force ourselves into the top-level package, even
# if we were loaded into a runtime-v1 per-app Anvil package
__package__ = "anvil.tables"
__name__ = "anvil.tables.query"


class _pattern_query(object):
    def __init__(self, pattern):
        self.pattern = pattern

    __hash__, __eq__ = _hash_wrapper("pattern")


class _value_query(object):
    def __init__(self, value):
        self.value = value

    __hash__, __eq__ = _hash_wrapper("value")


class _of_query(object):
    def __init__(self, *args, **kwargs):
        self.args = _make_refs(args)
        self.kwargs = _make_refs(kwargs)

    def __hash__(self):
        return hash(self.args + tuple(sorted(self.kwargs.items())))

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return self.args == other.args and self.kwargs == other.kwargs


#!defFunction(anvil.tables.query,_,pattern)!2: "Match values using a case-sensitive LIKE query, using the % wildcard character." ["like"]
@portable_class
class like(_pattern_query):
    pass


#!defFunction(anvil.tables.query,_,pattern)!2: "Match values using a case-insensitive ILIKE query, using the % wildcard character." ["ilike"]
@portable_class
class ilike(_pattern_query):
    pass


#!defFunction(anvil.tables.query,_,value)!2: "Match values greater than the provided value." ["greater_than"]
@portable_class
class greater_than(_value_query):
    pass


#!defFunction(anvil.tables.query,_,value)!2: "Match values less than the provided value." ["less_than"]
@portable_class
class less_than(_value_query):
    pass


#!defFunction(anvil.tables.query,_,value)!2: "Match values greater than or equal to the provided value." ["greater_than_or_equal_to"]
@portable_class
class greater_than_or_equal_to(_value_query):
    pass


#!defFunction(anvil.tables.query,_,value)!2: "Match values less than or equal to the provided value." ["less_than_or_equal_to"]
@portable_class
class less_than_or_equal_to(_value_query):
    pass


#!defFunction(anvil.tables.query,_,min,max,[min_inclusive=True],[max_inclusive=False])!2: "Match values between the provided min and max, optionally inclusive." ["between"]
def between(min, max, min_inclusive=True, max_inclusive=False):
    return all_of(
        greater_than_or_equal_to(min) if min_inclusive else greater_than(min),
        less_than_or_equal_to(max) if max_inclusive else less_than(max),
    )


#!defFunction(anvil.tables.query,_,query,[raw=False])!2: "Match values that match the provided full-text search query." ["full_text_match"]
@portable_class
class full_text_match(object):
    def __init__(self, query, raw=False):
        self.query = query
        self.raw = raw

    __hash__, __eq__ = _hash_wrapper("query", "raw")


#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match all query parameters given as arguments and keyword arguments" ["all_of"]
@portable_class
class all_of(_of_query):
    pass


#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match any query parameters given as arguments and keyword arguments" ["any_of"]
@portable_class
class any_of(_of_query):
    pass


#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match none of the query parameters given as arguments and keyword arguments" ["none_of"]
@portable_class
class none_of(_of_query):
    pass


#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match none of the query parameters given as arguments and keyword arguments" ["not_"]
not_ = none_of

#!defFunction(anvil.tables.query,_,rows)!2: "Define the number of rows that are fetched per round trip to the server." ["page_size"]
@portable_class
class page_size(object):
    def __init__(self, rows):
        self.rows = rows

    __hash__, __eq__ = _hash_wrapper("rows")


@portable_class("anvil.tables.fetch_only")
class fetch_only(object):
    def __init__(self, *only_cols, **linked_cols):
        spec = {}
        for col in only_cols:
            if not isinstance(col, str):
                raise TypeError("columns must be strings")
            spec[col] = True
        for col, only in linked_cols.items():
            if not isinstance(only, fetch_only):
                raise TypeError("keyword arguments must use q.fetch_only()")
            spec[col] = only.spec
        self.spec = spec

    def _hashable(self, val):
        if val is True:
            return val
        return self._as_tuple(val)

    def _as_tuple(self, spec):
        return tuple((col_name, self._hashable(val)) for col_name, val in sorted(spec.items()))

    def __hash__(self):
        return hash(self._as_tuple(self.spec))

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return other.spec == self.spec



import anvil

if not anvil.is_server_side():
    raise Exception("anvil.files cannot be imported on the client side.")

import anvil.server
import anvil.media
from anvil.tables.v2 import app_tables
from tempfile import gettempdir, mkdtemp
import os
import shutil
import sqlite3
from time import time, sleep
from contextlib import contextmanager
from uuid import uuid4
import logging
import sys

logger = logging.getLogger(__name__)

handler = logging.StreamHandler(stream=sys.stdout)
formatter = logging.Formatter(
    '[%(name)s %(levelname)s] %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)


def enable_debug_logging():
    logger.setLevel(logging.DEBUG)


MAX_REMOTE_METADATA_AGE = 5
DOWNLOAD_TIMEOUT = 5
FILES_TABLE = getattr(app_tables, "files") # TODO: Get table name from server config

class Files(object):

    def __init__(self):
        self._table_id = FILES_TABLE._id # TODO: Switch to .id when it works
        self._temp_dir = os.path.join(gettempdir(), "anvil-data-files")
        try:
            os.makedirs(self._temp_dir)
        except OSError:
            pass
        self._cache_dir = os.path.join(self._temp_dir, "table-%s" % self._table_id)
        
        self._db_path = os.path.join(self._temp_dir, "anvil-data-files-metadata.db")
        db = self._get_db()
        cur = db.cursor()
        cur.execute("CREATE TABLE IF NOT EXISTS tables (table_id PRIMARY KEY, last_fetched)")
        cur.execute("CREATE TABLE IF NOT EXISTS local_files (table_id, path, local_file_version, status, remote_file_version, last_touched, PRIMARY KEY (table_id, path) on conflict fail)")
        cur.execute("CREATE TABLE IF NOT EXISTS remote_files (table_id, path, file_version, PRIMARY KEY (table_id, path))")
        db.commit()

    def _get_db(self):
        db = sqlite3.connect(self._db_path)
        db.row_factory = sqlite3.Row
        return db

    def download(self, db, remote_file_metadata):
        path = remote_file_metadata['path']
        local_file_path = os.path.join(self._cache_dir, path)
        file_version = remote_file_metadata['file_version']
        logger.debug("Downloading %s" % path)
        db.execute("UPDATE local_files SET status='DOWNLOADING', local_file_version = NULL, remote_file_version = ?, last_touched = ? WHERE table_id = ? AND path = ?",
                    [file_version, time(), self._table_id, path])
        db.commit()
        remote_file_row = FILES_TABLE.get(path=path)
        d = mkdtemp()
        try:
            tempname = os.path.join(d, '%s.anvildownload' % path)
            try:
                os.makedirs(os.path.dirname(tempname))
            except OSError:
                pass
            logger.debug("Downloading to %s" % tempname)
            with open(tempname, 'wb+') as f:
                # TODO: Periodically update last_touched while downloading
                if remote_file_row['file']:
                    f.write(remote_file_row['file'].get_bytes())

            try:
                os.makedirs(os.path.dirname(local_file_path))
            except OSError:
                pass
            os.rename(tempname, local_file_path)
        finally:
            shutil.rmtree(d, ignore_errors=True)

        logger.debug("Downloaded %s to %s" % (path, local_file_path))

        db.execute("UPDATE local_files SET status='PRESENT', local_file_version = ?, last_touched = ? WHERE table_id = ? AND path = ?",
                    [file_version, time(), self._table_id, path])
        db.commit()

    def upload(self, db, path):
        # TODO: Cope with new files?
        file_row = FILES_TABLE.get(path=path)
        local_file_path = os.path.join(self._cache_dir, path)
        logger.debug("Uploading %s" % path)
        db.execute("UPDATE local_files SET status='UPLOADING', last_touched=? WHERE table_id=? AND path=?", [time(), self._table_id, path])
        db.commit()
        file_row['file'] = anvil.media.from_file(local_file_path) # TODO: MIME type? Store in metadata?
        new_file_version = str(uuid4()) # TODO: Proper content hash
        file_row['file_version'] = new_file_version
        logger.debug("Uploaded %s" % path)
        db.execute("UPDATE remote_files SET file_version=? WHERE table_id=? AND path=?", [new_file_version, self._table_id, path])
        db.commit()
        db.execute("UPDATE local_files SET status='PRESENT', last_touched=?, local_file_version=?, remote_file_version=? WHERE table_id=? AND path=?", [time(), new_file_version, new_file_version, self._table_id, path])
        db.commit()


    def __getitem__(self, path):
        db = self._get_db()
        cur = db.cursor()
        local_path = os.path.join(self._cache_dir, path)

        # Fetch and possibly upsert remote table metadata into local DB.
        cur.execute("SELECT * FROM tables WHERE table_id = ?", [self._table_id])
        table_metadata = cur.fetchone()
        if table_metadata is None or time() - table_metadata['last_fetched'] > MAX_REMOTE_METADATA_AGE:
            logger.debug("Fetching remote file metadata")
            db.execute("INSERT OR REPLACE INTO tables (table_id, last_fetched) VALUES (?, ?)", [self._table_id, time()])
            db.execute("DELETE FROM remote_files WHERE table_id = ?", [self._table_id])
            for file in FILES_TABLE.search():
                db.execute("INSERT INTO remote_files (table_id, path, file_version) VALUES (?, ?, ?)", [self._table_id, file['path'], file['file_version']])
            db.commit()

        # Fetch remote file metadata
        path = path.strip("/")
        path_like = "%s/%%" % path

        cur.execute("SELECT * FROM remote_files WHERE table_id = ? AND (path LIKE ? or path = ?) ORDER BY path = ?", [self._table_id, path_like, path, path])
        remote_files_metadata = cur.fetchall()

        if not remote_files_metadata:
            # Whatever we were looking for, it doesn't exist remotely. Remove matching local file/directory
            # Path might be a file:
            try:
                os.remove(os.path.join(self._cache_dir, path))
            except OSError:
                pass
            # Path might be a directory
            shutil.rmtree(os.path.join(self._cache_dir, path), ignore_errors=True)

            raise Exception("File not found: %s" % path)
        else:
            # We might have matched multiple remote files (if path is a directory)
            is_folder = len(remote_files_metadata) > 1
            for remote_file_metadata in remote_files_metadata:
                logger.debug("Loading file: %s" % remote_file_metadata['path'])
                if is_folder and remote_file_metadata['path'] == path:
                    logger.warning("Found file with same name as folder: %s" % path)
                    continue

                try:
                    db.execute("INSERT OR FAIL INTO local_files (table_id, path, remote_file_version, status) VALUES (?,?,?,'DOWNLOADING')", [self._table_id, remote_file_metadata['path'], remote_file_metadata['file_version']])
                    db.commit()
                    # We managed to insert the metadata, which means it didn't already exist. Download the file.
                    self.download(db, remote_file_metadata)

                except sqlite3.IntegrityError:
                    # The metadata exists, so fetch it.
                    cur.execute("SELECT * FROM local_files WHERE table_id = ? AND path = ?", [self._table_id, remote_file_metadata['path']])
                    local_file_metadata = cur.fetchone()

                    if local_file_metadata['status'] == 'DOWNLOADING':
                        logger.debug("File is already downloading")
                        while time() - local_file_metadata['last_touched'] < DOWNLOAD_TIMEOUT:
                            logger.debug("Waiting for download to finish")
                            # We haven't been downloading long. Wait and see.
                            sleep(0.5)
                            cur.execute("SELECT * FROM local_files WHERE table_id = ? AND path = ?", [self._table_id, remote_file_metadata['path']])
                            local_file_metadata = cur.fetchone()
                            if local_file_metadata['status'] == 'PRESENT':
                                logger.debug("Download finished elsewhere")
                                break
                        else:
                            # The download has probably stalled. Try again.
                            logger.debug("Previous download timed out")
                            self.download(db, remote_file_metadata)

                    elif local_file_metadata['status'] == 'UPLOADING':
                        logger.debug("Already uploading")
                        # TODO: Something?
                    elif local_file_metadata['status'] == 'PRESENT':
                        if local_file_metadata['local_file_version'] == remote_file_metadata['file_version']:
                            logger.debug("Found local cache")
                        else:
                            logger.debug("Local file cache has invalid hash.")
                            self.download(db, remote_file_metadata)

                    else:
                        raise Exception("Invalid local file status: %s" % local_file_metadata['status'])

            logger.debug("")
            logger.debug("TABLES")
            cur.execute("SELECT * FROM tables")
            for r in cur.fetchall():
                logger.debug(dict(r))
            logger.debug("REMOTE FILES")
            cur.execute("SELECT * FROM remote_files")
            for r in cur.fetchall():
                logger.debug(dict(r))

            logger.debug("LOCAL FILES")
            cur.execute("SELECT * FROM local_files")
            for r in cur.fetchall():
                logger.debug(dict(r))
            logger.debug("")

            return local_path

    def editing(self, path):
        local_path = self[path]
        db = self._get_db()
        table_id = self._table_id
        upload = self.upload

        class Editing():
            def __enter__(self):
                return local_path

            def __exit__(self, exc_type, exc_val, exc_tb):
                upload(db, path)

        return Editing()

    @contextmanager
    def open(self, path, mode='r'):
        with open(data_files[path], mode) as f:
            yield f
        if "w" in mode or "a" in mode or "+" in mode or "x":
            self.upload(self._get_db(), path)

data_files = Files()




import anvil.server

def get_user_email():
    return anvil.server.call("anvil.private.facebook.auth.get_user_email")

def get_user_id():
    return anvil.server.call("anvil.private.facebook.auth.get_user_id")

def get_user_access_token():
    return anvil.server.call("anvil.private.facebook.auth.get_user_access_token")

def login(*a, **k):
	raise Exception("facebook.auth.login cannot be called from a server module")



# For the sake of a soft roll-out, we don't want to load v2 code implicitly
# from anvil.tables.query, but that module needs access to `make_refs` if we're
# using v2. So we inject it (only) when v2 loads.

from .. import query
from . import _refs

query._make_refs = _refs.make_refs



import anvil.google.auth
import anvil.regex
import anvil.server as rpc

def index_to_col(idx):
    assert type(idx) is int, "Expected an integer for column index not " + type(idx).__name__
    idx -= 1
    if idx < 26:
        return chr(65 + (idx % 26))
    else:
        idx -= 26
        return chr(65 + (idx // 26)) + chr(65 + (idx % 26))

class ApiItem(object):
    def __new__(cls, obj=None, other=None):
        # use __new__ rather than __init__ since __getattr__ relies on _obj existing
        # otherwise we can end up in a recursion error during deserialization where the _obj is yet to exist
        # and we attempt to getattr the __deserialize__ method
        self = object.__new__(cls)
        self._obj = obj or {}
        self._other = other or {}
        return self

    def __getitem__(self, item_name):
        return self._obj[item_name]

    def __getattr__(self, attr_name):
        try:
            return self._obj[attr_name]
        except KeyError:
            raise AttributeError("Could not find attribute: %s" % attr_name)

    def __dir__(self):
        return object.__dir__(self) + list(self._obj.keys())

@anvil.server.portable_class
class Cell(ApiItem):
    #!defAttr()!1: {name:"row",type:"number",description:"This cell's row index (starting from 1)"}
    #!defAttr()!1: {name:"col",type:"number",description:"This cell's column index (starting from 1)"}
    #!defAttr()!1: {name:"value",type:"string",description:"The value in this cell"}
    #!defAttr()!1: {name:"input_value",type:"string",description:"The value that was entered into the cell"}
    def __setattr__(self, name, value):
        if name == "value" or name == "input_value":

            val = rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"], value)

            self._obj["input_value"] = str(value)
            self._obj["value"] = val

            return val
        else:
            object.__setattr__(self, name, value)

    def __repr__(self):
        return "<Google Worksheet Cell: %r>" % self._obj["value"]
#!defClass(anvil.google.sheets,#Cell)!:


def _old_field_transform(new_field_name):
    # for backwards compatability before v4
    return anvil.regex.replace(new_field_name, "[^A-Za-z0-9\\-]", "").lower()

@anvil.server.portable_class
class Row(ApiItem):
    def __getitem__(self, key):
        return self._other["data"][self._other["fields_old"].get(key, key)]

    def __setitem__(self, key, value):
        key = self._other["fields_old"].get(key, key)
        if key in self._other["data"]:
            rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"].narrow([index_to_col(self._other["fields"].index(key) + 1)]), value)
        else:
            raise KeyError(key)

    #!defMethod(_)!2: "Delete this row from the worksheet. (This will cause data in subsequent rows to shift up)" ["delete"]
    def delete(self):
        rpc.call("anvil.private.google.sheets.v4.delete_row", self._other["capability"])

    def __iter__(self):
        return iter(self._other["data"])

    def keys(self):
        return self._other["data"].keys()

    def __repr__(self):
        return "<Google Worksheet Row: %s>" % self._other["data"]
#!defClass(anvil.google.sheets,#Row)!:


class RowIterator(object):
    def __init__(self, cap, query_dict):
        self._cap = cap

        self._fields, rows, self._next_start, self._done = rpc.call(
            "anvil.private.google.sheets.v4.list_initial_rows", cap, query_dict
        )
        self._query_list = [query_dict.get(f) for f in self._fields]
        self._fields_old = {_old_field_transform(f): f for f in self._fields}
        self._iterator = iter(rows)

    def _load_next_page(self):
        rows, self._next_start, self._done = rpc.call(
            "anvil.private.google.sheets.v4.list_more_rows", self._cap, self._query_list, self._next_start
        )
        self._iterator = iter(rows)

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            try:
                r = next(self._iterator)
                return Row({}, {
                    "data": {field: (r["data"][i] if i < len(r["data"]) else "") for i, field in enumerate(self._fields)},
                    "capability": self._cap.narrow([r["row"]]),
                    "fields": self._fields,
                    "fields_old": self._fields_old
                })
            except StopIteration:
                if self._done:
                    raise
                else:
                    self._load_next_page()

    next = __next__


class RowList(object):
    def __init__(self, cap, query_dict):
        self._cap = cap
        self._query_dict = query_dict
        self._it = RowIterator(cap, query_dict)

    def __iter__(self):
        # TODO this won't work multiple times client side - we'll get a SuspensionError on the second use of this iterator
        if not self._it:
            return RowIterator(self._cap, self._query_dict)
        try:
            return self._it
        finally:
            self._it = None


@anvil.server.portable_class
class Worksheet(ApiItem):

    #!defAttr()!1: {name:"title",type:"string",description:"The title of this worksheet"}
    #!defAttr()!1: {name:"row_count",type:"number",description:"The number of rows in this worksheet"}
    #!defAttr()!1: {name:"column_count",type:"number",description:"The number of columns in this worksheet"}
    #!defAttr()!1: {name:"fields",type:"list",description:"The fields in this worksheet (ie the column headers, or the values in the first row)"}
    #!defAttr()!1: {name:"rows",pyType:"list(anvil.google.sheets.Row instance)",description:"The rows in this worksheet (excluding the header)"}
    #!defAttr()!1: {name:"cells",pyType:"list(anvil.google.sheets.Cell instance)",description:"A list of all the cells in this worksheet"}
    def __getattr__(self, name):
        if name == "fields":
            return self.get_fields()
        elif name == "rows":
            return list(self.list_rows())
        elif name == "cells":
            return list(self.list_cells())
        else:
            return ApiItem.__getattr__(self, name)

    #!setItemType(anvil.google.sheets.Cell instance)!:
    def __getitem__(self, cell_tuple):
        row, col = cell_tuple

        return self.get_cell(row, col)

    def get_fields(self):
        cs = self._obj["column_count"]
        return [c.value for c in self.list_cells(1, 1, 1, cs)]

    #!defMethod(list[anvil.google.sheets.Row instance],**query)!2: "List rows in this worksheet, optionally restricting to rows with the specified column values specified as keyword arguments" ["list_rows"]
    def list_rows(self, **query):
        return RowList(self._other["capability"], query)

    #!defMethod(anvil.google.sheets.Row instance,**fields)!2: "Add a row to the end of the worksheet, specifying values for columns as keywords arguments" ["add_row"]
    def add_row(self, **kwargs):

        fields = self.get_fields()
        fields_old = {_old_field_transform(field): field for field in fields}

        values = {fields_old.get(kw, kw): str(v) for kw, v in kwargs.items()}
        data = [values.get(field, "") for field in fields]
        added_row_index = rpc.call("anvil.private.google.sheets.v4.add_row", self._other["capability"], data)

        return Row({}, {
            "data": {field: value for field, value in zip(fields, data)},
            "capability": self._other["capability"].narrow([int(added_row_index)]),
            "fields": fields,
            "fields_old": fields_old
        })

    #!defMethod(list[anvil.google.sheets.Cell instance],[min_row=],[max_row=],[min_col=],[max_col=])!2: "List cells in the worksheet, optionally specifying a region" ["list_cells"]
    def list_cells(self, min_row=None, max_row=None, min_col=None, max_col=None):
        if min_row is max_row is min_col is max_col is None:
            range_str = ""
            min_row = min_col = 1
        else:
            min_col_str = index_to_col(min_col)
            max_col_str = index_to_col(max_col)

            range_str = min_col_str + str(min_row) + ":" + max_col_str + str(max_row)

        cells = rpc.call(
            "anvil.private.google.sheets.v4.list_cells",
            self._other["capability"],
            range_str,
            [min_row, min_col, max_row, max_col],
        )
        return [
            Cell(c, {"capability": self._other["capability"].narrow([c["row"], index_to_col(c["col"])])}) for c in cells
        ]

    #!defMethod(anvil.google.sheets.Cell instance,row,col)!2: "Get a particular cell from the spreadsheet" ["get_cell"]
    def get_cell(self, row, col):
        cells = self.list_cells(row, row, col, col)
        if len(cells) > 0:
            return cells[0]
        return None

    def __repr__(self):
        return "<Google Worksheet: %r>" % self._obj["title"]

    def __dir__(self):
        return ApiItem.__dir__(self) + ["cells", "fields", "rows"]
#!defClass(anvil.google.sheets,#Worksheet)!:


@anvil.server.serializable_type
class Spreadsheet(ApiItem):
    #!setItemType(anvil.google.sheets.Worksheet instance)!:
    def __getitem__(self, name):
        if isinstance(name, str):
            for w in self.list_worksheets():
                if w.title == name:
                    return w
            raise KeyError("Spreadsheet contains no worksheet '%s'" % name)
        else:
            return list(self.list_worksheets())[name]

    #!defAttr()!1: {name:"worksheets",pyType:"list(anvil.google.sheets.Worksheet instance)",description:"The worksheets in this spreadsheet."}
    #!defAttr()!1: {name:"title",type:"string",description:"The title of this spreadsheet."}
    #!defAttr()!1: {name:"id",type:"string",description:"The ID of this spreadsheet in Google Drive"}
    def __getattr__(self, name):
        if name in self._obj:
            return self._obj[name]
        elif name == "worksheets":
            return self
        else:
            raise AttributeError("Could not find attribute: %s" % name)

    def __iter__(self):
        return iter(self.list_worksheets())

    def __len__(self):
        return len(list(self.list_worksheets()))

    #!defMethod(list[anvil.google.sheets.Worksheet instance])!2: "Get a list of all worksheets in this spreadsheet" ["list_worksheets"]
    def list_worksheets(self):
        return [Worksheet({
            "title": w["title"],
            "row_count": w["gridProperties"]["rowCount"],
            "column_count": w["gridProperties"]["columnCount"]
        }, {
            "capability": self._other["capability"].narrow([[w["title"], w["sheetId"]]])
        }) for w in self._other["worksheets"]]


    def __repr__(self):
        return "<Google Sheet: %r>" % self._obj["title"]
#!defClass(anvil.google.sheets,#Spreadsheet)!:


Sheet = Spreadsheet  # Alias for backwards-compatibility


def login():
    return anvil.google.auth.login(["https://www.googleapis.com/auth/spreadsheets"])


def get_sheet(id, creds):
    s = Spreadsheet(**rpc.call("anvil.private.google.sheets.v4.get_sheet", id, creds))
    return s


"""
id: drive_sheets
docs_url: /docs/integrations/google/google-drive#google-sheets
title: Google Sheets
description: |
  ```python
  db = app_files.my_sheet
  ```

  Google Sheets make ideal databases for Anvil apps. You can access sheets using Google Drive in the same way you access files.

  ```python
  ws = db["Sheet 1"]
  ```

  Once you have an object representing a Google Sheet, you can look up its individual worksheets. (It is also possible to look up a worksheet by number, eg `db[0]`, or obtain a list with `list(db.worksheets)`.)

  ```python
  print db["Sheet 1"].fields
  # Expect output similar to: '[name, age]'
  ```

  You should set up your Sheet using [Google Drive](https://drive.google.com) so that the column names are in the first row. Once your sheet is set up, you can get a list of the fields in a worksheet using the `fields` attribute.

  Fields will always be in lower case (even if your column headings use capital letters).

  ```python
  # Assuming columns "name" and "age"
  for r in db["Sheet 1"].rows:
    print "%s is %s years old" % (r["name"], r["age"])
  ```

  You can access rows of data like so:

  ```python
  # Convert age to dog years:
  for r in db["Sheet 1"].rows:
    r["age"] = 7 * int(r["age"])
  ```

  You can update them as well as reading them. You can assign numbers or strings,
  but the data you read from a row will always be a string:

  ```python
  ws = db["Sheet 1"]
  row = ws.add_row(name="Bob", age=56)
  ```

  You can add rows to your spreadsheet using the `add_row(...)` method:

  (Note: `add_row()` normally returns the new created row. However, if you
  attempt to insert an empty row, it will return `None`.)

  An Exception will be raised if you try to set a non-existent field.

  ```python
  row.delete()
  ```

  You can delete a row using the `delete()` method.

  Beware: If you delete a row, other rows in the worksheet may shift unpredictably. If you delete a row, you should not re-use any other `row` objects from that worksheet. Call `list_rows()` again to find the rows you want again.

  ```python
  ## Display the calculated value of
  ## the cell in row 1, col 2

  print ws[1,2].value

  ## Display the formula in cell (1,2)

  print ws[1,2].input_value

  ## Set the value of a cell

  ws[3,4].value = "Some text"
  ```

  It is also possible to access individual cells of your spreadsheet.

"""



import sys

__author__ = 'Meredydd Luff <meredydd@anvil.works>'

_warnings = {}


class LiveObject():
    def __init__(self, spec):
        self._spec = spec

    def __repr__(self):
        return "<LiveObject: " + self._spec["backend"] + ">"

    def __eq__(self, other):
      return isinstance(other, LiveObject) and other._spec["id"] == self._spec["id"] and other._spec["backend"] == self._spec["backend"]

    def __ne__(self, other):
      return not self.__eq__(other)

    def __hash__(self):
        return hash((self._spec["id"], self._spec["backend"]))


# Hack
def _get_live_object_id(lo):
    if isinstance(lo, LiveObject):
        return lo._spec["id"]
    elif hasattr(lo, "get_id"):
        if _warnings.get("_get_live_object_id") is None:
            _warnings["_get_live_object_id"] = True
            print("Deprecated: _get_live_object_id is no longer required - call row.get_id() instead.")
        return lo.get_id()
    else:
        raise Exception("Not a LiveObject")

def _clear_live_object_caches(lo):
    if isinstance(lo, LiveObject):
        lo._spec["itemCache"] = {}
        lo._spec["iterItems"] = {}
    else:
        raise Exception("Not a LiveObject")

class Media(object):
    @property
    def url(self):
        return self.get_url()
    @property
    def content_type(self):
        return self.get_content_type()
    @property
    def name(self):
        return self.get_name()
    @property
    def length(self):
        return self.get_length()
    
    def get_content_type(self):
        return None

    def get_url(self, download=True):
        return None

    def get_bytes(self):
        raise Exception("get_bytes() not implemented")

    def get_length(self):
        return len(self.get_bytes())

    def get_name(self):
        return None


_byte_string_type = bytes if sys.version_info >= (3,) else basestring
_unicode_string_type = str if sys.version_info >= (3,) else unicode


class BlobMedia(Media):
    def __init__(self, content_type, content, name=None):
        self._content_type = content_type
        if not isinstance(content, _byte_string_type):
            raise TypeError("BlobMedia content must be a byte string.")
        elif isinstance(content, _unicode_string_type):
            content = content.encode("utf-8")
        self._bytes = content
        self._name = name

    def get_content_type(self):
        return self._content_type

    def get_bytes(self):
        return self._bytes

    def get_name(self):
        return self._name

    def __repr__(self):
        return "BlobMedia[%s,%d bytes%s]" % (self._content_type, len(self._bytes), (",name=" + self._name if self._name is not None else ""))

DataMedia = BlobMedia


class URLMedia(Media):
    def __init__(self, url):
        self._url = url
        self._fetched = None

    def _fetch(self):
        import anvil.server
        if self._fetched is None:
            self._fetched = anvil.server.call("anvil.private.http.request", url=self._url)['content']
        return self._fetched

    def get_url(self, download=True):
        return self._url

    def get_bytes(self):
        return self._fetch().get_bytes()

    def get_content_type(self):
        return self._fetch().content_type


def create_lazy_media(mime_type, server_function_name, *args, **kwargs):
    import anvil.server
    return anvil.server.call("anvil.private.mk_LazyMedia", mime_type, server_function_name, *args, **kwargs)


def plot_image():
    import anvil.mpl_util
    return anvil.mpl_util.plot_image()


def is_server_side():
    return True


def _get_service_client_config(path):
    import anvil.server
    # TODO we could cache the return value - but we don't use it anywhere
    return anvil.server.call("anvil.private.get_client_config", path)


class _AppInfo:
    class _Environment:
        def __init__(self, description=None, tags=[], **kwargs):
            self.__dict__.update(name=description, tags=tags, **kwargs)

        def __setattr__(self, key, value):
            raise AttributeError("This object is read-only")
        
        def __repr__(self):
            info = ", ".join("{}={!r}".format(key, val) for key, val in self.__dict__.items())
            return "Environment({})".format(info)

    def __init__(self, id, branch):
        self.__dict__.update(id=id, branch=branch, environment=_AppInfo._Environment())

    def __setattr__(self, key, value):
        raise AttributeError("This object is read-only")

    def _setup(self, environment={}, **kwargs):
        self.__dict__.update(kwargs)
        self.__dict__['environment'] = _AppInfo._Environment(**environment)


app = _AppInfo(None, None)

from ._components import *



import os.path, re, sys, subprocess, shlex, zipfile, progressbar

if sys.version_info < (3,0,0):
    import urllib
    _urlretrieve = urllib.urlretrieve
else:
    import urllib.request
    _urlretrieve = urllib.request.urlretrieve


def launch():
    jar_path = find_or_download_app_server()
    java_args = ["java", "-jar", jar_path]
    if sys.executable:
        os.environ["PYTHON_INTERPRETER"] = sys.executable
    try:
        return_code = subprocess.call(java_args + sys.argv[1:])
    except OSError:
        print("Failed to launch: java -jar {}".format(jar_path))
        print("The Anvil App Server requires a Java Virtual Machine (JVM) installed.")
        print("Install with 'apt install openjdk-8-jdk' or 'brew install openjdk', or")
        print("download it from https://adoptopenjdk.net")
        return_code = 1
    except KeyboardInterrupt:
        return_code = 0
    sys.exit(return_code)


def create_app():
    # User invokes script with
    # `create-anvil-app <template_name> <directory_name>`

    # Unzip the zip file corresponding to the template_name into a directory called directory_name
    TEMPLATES = {
      'hello-world': {'filename':'Default.zip','description':'A simple Anvil app in Material Design theme'},
      'blank': {'filename':'Blank.zip','description':'A blank Anvil app in Material Design theme'},
      'todo-list': {'filename':'TodoList.zip','description':'A simple To-do list (CRUD) app in Material Design theme'},
    }

    def describe_templates():
        for name, info in TEMPLATES.items():
            print("    {}: {}".format(name, info['description']))
        print("\nFor example:\n\n   create-anvil-app todo-list ToDoList\n")


    # Condition on template_name being correct || Give helpful error messages if the template_name provided doesn't exist

    # Verify length of sys.argv first
    if len(sys.argv) != 3:
        print("\nPlease specify both a template app and directory_name, in the format:\n\n    create-anvil-app <template_name> <directory_name>\n")
        print("The template_name can be:\n")
        describe_templates()
        sys.exit(1)

    template_input = sys.argv[1]

    if template_input not in TEMPLATES:
        # Tells user what templates are available
        print("\nTemplate '{}'' does not exist. Please choose from one of the following:\n".format(template_input))
        describe_templates()
        sys.exit(1)

    # Find zip file corresponding to name
    filename = TEMPLATES[template_input]['filename']
    directory_name = sys.argv[2]

    if not re.match('^[A-Za-z_][A-Za-z0-9_]*$', directory_name):
        print("\nThe directory '{}' is not a valid Python package name.".format(directory_name))
        print("Packages must start with a letter or underscore, and be made up of letters, numbers and underscores.")
        sys.exit(1)

    package_dir = os.path.dirname(__file__)
    package_dir_path = os.path.join(package_dir, filename)

    # Unzip file
    with zipfile.ZipFile(package_dir_path, 'r') as zip_ref:
        zip_ref.extractall(directory_name)

    print("""
Congratulations! Your new Anvil app is created, using the '{}'
template. To serve this app, run:

anvil-app-server --app {}

""".format(template_input, directory_name))


def psql():
    data_path = ".anvil-data"
    if len(sys.argv) > 2:
        print("Usage: psql-anvil-app-server [<data-dir>]")
        sys.exit(1)
    elif len(sys.argv) == 2:
        data_path = sys.argv[1]

    print("Looking for postmaster.opts in {} ...".format(data_path+os.path.sep))
    print("(Use 'psql-anvil-app-server <data-dir>' to look somewhere else.)")

    opts_path = os.path.join(data_path,"db","postmaster.opts")
    try:
        with open(opts_path, "r") as f:
            opts = f.read()
    except OSError:
        print("Could not open {}. Is {} an Anvil App Server data directory?".format(opts_path, data_path))
        sys.exit(1)

    port = None
    last = None
    for opt in shlex.split(opts):
        if last == "-p":
            port = int(opt)
            break
        else:
            last = opt

    if port is None:
        print("Could not determine Postgres port from {}.\nIs {} an Anvil App Server data directory?".format(opts_path, data_path))
        sys.exit(1)

    pw_path = os.path.join(data_path, "postgres.password")
    try:
        with open(pw_path, "r") as f:
            os.environ["PGPASSWORD"] = f.read()
    except OSError:
        print("Could not open {}. Is {} an Anvil App Server data directory?".format(pw_path, data_path))
        sys.exit(1)

    has_pgcli = False
    try:
        import shutil
        if shutil.which("pgcli"):
            has_pgcli = True
    except ImportError:
        pass
    if has_pgcli:
        print("Using 'pgcli':")
        print("pgcli 'host=localhost port={} user=postgres dbname=postgres options=--search-path=app_tables'".format(port))
        os.system("pgcli 'host=localhost port={} user=postgres dbname=postgres options=--search-path=app_tables'".format(port))
    else:
        print("psql 'host=localhost port={} user=postgres dbname=postgres options=--search-path=app_tables'".format(port))
        os.system("psql 'host=localhost port={} user=postgres dbname=postgres options=--search-path=app_tables'".format(port))


def find_or_download_app_server(): # Work out whether we already have the server JAR file. It could be in the package itself, or in ~/.anvil
    package_dir = os.path.dirname(__file__)
    home_dir = os.path.expanduser("~")
    server_jar_name = "anvil-app-server.20220602-121046.jar"

    package_dir_path = os.path.join(package_dir, server_jar_name)
    home_dir_path = os.path.join(home_dir, ".anvil", server_jar_name)

    if os.path.isfile(package_dir_path):
        print("Found Anvil App Server JAR in package directory")
        return package_dir_path
    if os.path.isfile(home_dir_path):
        print("Found Anvil App Server JAR in ~/.anvil directory")
        return home_dir_path

    # If we don't have it, download it to the package if possible, or to ~/.anvil if not possible.
    progress = { "bar": None, "downloaded": 0 }
    def show_progress(count, block_size, total_size):
        if progress['bar'] is None:
            progress['bar'] = progressbar.DataTransferBar(max_value=total_size)

        progress['downloaded'] = count * block_size
        if progress['downloaded'] >= total_size:
            progress['bar'].finish()
            progress['bar'] = None
            progress['downloaded'] = 0
        else:
            progress['bar'].update(progress['downloaded'])


    url = "https://anvil-public-assets.s3.eu-west-2.amazonaws.com/app-server/" + server_jar_name
    try:
        print("Downloading Anvil App Server JAR to package directory")
        _urlretrieve(url, package_dir_path, show_progress)
        print("Downloaded Anvil App Server JAR to package directory")
        return package_dir_path
    except Exception:
        print("Failed to download App Server to package directory. Retrying in ~/.anvil")
        _urlretrieve(url, home_dir_path, show_progress)
        print("Downloaded Anvil App Server JAR to ~/.anvil directory")
        return home_dir_path


if __name__ == "__main__":
    launch()



import anvil.server

import json as json_mod

class HttpErrorStatus(Exception):
    "Represents an HTTP error response (eg 404 Not Found)"
    def __init__(self, status, content, message=None):
        self.status = status
        self.content = content
        if message is None:
            message = "HTTP error %s" % status
        Exception.__init__(self, message)
#!defAttr()!1: {name: "status", type: "number", description: "The numeric HTTP status error (eg 404 for \"not found\").\n\nStatus will be 0 for errors that prevent the request completing at all (eg cross-origin policy in the browser)."}
#!defAttr()!1: {name: "content", pyType: "anvil.Media instance", description: "The content returned by the request (eg the body of a 404 response)"}
#!defClass(anvil.http,HttpError)!0:


# Backward compatibility
HttpError = HttpErrorStatus


class HttpRequestFailed(anvil.server.AnvilWrappedError):
    "Represents an HTTP request that failed to complete entirely"
    pass


anvil.server._register_exception_type("anvil.http.HttpRequestFailed", HttpRequestFailed)

#!defFunction(anvil.http,_,url,[method="GET"],[data=None],[json=False],[headers=None],[username=None],[password=None], [timeout=None])!2: "Make an HTTP request to the specified URL.\n\nIf json=True, the response is parsed into Python objects (dicts/lists/etc), and 'data' is JSON-encoded before sending.\n\n'headers' can be a dict of strings to set HTTP headers.\nIf specified, 'username' and 'password' will be used to perform HTTP Basic authentication." ["request"]
def request(url, method='GET', data=None, headers=None, username=None, password=None, json=False, timeout=None):
    if headers is None:
        headers = {}
    if json and data is not None:
        data = json_mod.dumps(data)
        headers["content-type"] = "application/json"
    if timeout is not None:
        if not isinstance(timeout, (int, float)):
            raise TypeError("timeout must be a number")
        timeout = timeout * 1000

    resp = anvil.server.call(
        "anvil.private.http.request",
        url=url,
        method=method,
        data=data,
        headers=headers,
        username=username,
        password=password,
        timeout=timeout,
    )
    # Parse JSON if we have it

    if json:
        try:
            b = resp["content"].get_bytes()
            # Ew. microjson barfs if we give it a "unicode" object.
            # We only use microjson in the old 2.7 sandbox, so
            # this can go away soon.
            if not isinstance(b, str):
                b = b.decode()
            resp["content"] = json_mod.loads(b)
        except:
            raise HttpError(resp["status"], resp["content"], "Invalid JSON in response")

    if resp["status"] < 200 or resp["status"] > 299:
        raise HttpError(resp["status"], resp["content"])

    return resp["content"]


#!defFunction(anvil.http,_,string_to_encode)!2: "URL-encode a string" ["url_encode"]
def url_encode(string_to_encode):
    return anvil.server.call("anvil.private.http.url_encode", string_to_encode)

#!defFunction(anvil.http,_,string_to_encode)!2: "URL-decode a string. Raises UrlEncodingError on failure." ["url_decode"]
def url_decode(string_to_decode):
    return anvil.server.call("anvil.private.http.url_decode", string_to_decode)


#!defClass(anvil.http,UrlEncodingError)!0:
class UrlEncodingError(anvil.server.AnvilWrappedError):
    pass


anvil.server._register_exception_type("anvil.http.UrlEncodingError", UrlEncodingError)



import json, struct, threading

try:
    bytes
except NameError:
    bytes = str


class MessagePipe(object):
    def __init__(self, pipe):
        self.pipe = pipe
        self.lock = threading.RLock()

    def send(self, message, bindata=None):
        encoded_message = json.dumps(message).encode()
        self.send_encoded(encoded_message, bindata)

    def send_encoded(self, encoded_message, bindata=None):
        assert(type(encoded_message) is bytes)
        with self.lock:
            l = len(encoded_message)
            assert(l < 2**32)
            if bindata is not None:
                self.pipe.write(struct.pack("=?I", True, l) + encoded_message + struct.pack("I", len(bindata)))
                self.pipe.write(bindata)
            else:
                self.pipe.write(struct.pack("=?I", False, l) + encoded_message)
            self.pipe.flush()

    def _fully_receive(self, l):
        if l == 0:
            return b''
        s = self.pipe.read(l)
        while len(s) < l:
            r = self.pipe.read(l-len(s))
            if len(r) == 0:
                raise EOFError
            s += r
        return s

    def receive(self):
        has_bindata, msg_len = struct.unpack("=?I", self._fully_receive(5))
        message = json.loads(self._fully_receive(msg_len))
        bindata = None
        if has_bindata:
            bindata_len, = struct.unpack("I", self._fully_receive(4))
            bindata = self._fully_receive(bindata_len)

        return message, bindata



import anvil.server
from anvil.server import Capability

from .._base_classes import SearchIterator as BaseSearchIterator
from ._constants import CAP_KEY, SERVER_PREFIX, SHARED_DATA_KEY
from ._row import Row
from ._utils import check_serialized, init_spec_rows, init_view_data, merge_row_data, validate_cap

PREFIX = SERVER_PREFIX + "search."


class PartialSearchIter(object):
    def __init__(self, s, slice_):
        self._view_key = s._view_key
        self._table_id = s._table_id
        self._cap = s._cap
        self._idx = slice_.start or 0
        self._step = slice_.step or 1
        self._stop = slice_.stop
        row_ids, cap_next = s._row_ids, s._cap_next
        if row_ids is None:
            # this can happen in deserialization from untrusted/None transmited data
            row_ids, cap_next = [], s._cap
        assert cap_next is None or type(cap_next) is Capability
        self._reset(row_ids, cap_next, s._table_data)

    def _reset(self, row_ids, cap_next, table_data):
        if self._stop is not None and len(row_ids) > self._stop:
            row_ids, cap_next = row_ids[: self._stop], None
        self._row_ids = row_ids
        self._cap_next = cap_next
        self._table_data = table_data

    def _iter_next_page(self):
        if self._cap_next is None:
            raise StopIteration

        num_row_ids = len(self._row_ids)
        self._idx -= num_row_ids
        if self._stop is not None:
            self._stop -= num_row_ids

        row_ids, cap_next, table_data = anvil.server.call(PREFIX + "next_page", self._cap_next)

        self._reset(row_ids, cap_next, table_data)
        return self.__next__()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            row_id = self._row_ids[self._idx]
        except IndexError:
            return self._iter_next_page()
        self._idx += self._step
        return Row._create_from_trusted(self._view_key, self._table_id, row_id, self._table_data)

    next = __next__


@anvil.server.portable_class
class SearchIterator(BaseSearchIterator):
    @classmethod
    def _create(cls, view_key, table_id, row_ids, cap, cap_next, table_data):
        self = object.__new__(cls)
        assert cap_next is None or type(cap_next) is Capability
        self._view_key = view_key
        self._table_id = table_id
        self._row_ids = row_ids
        self._cap = cap
        self._cap_next = cap_next
        self._table_data = table_data
        self._from_serialize = False
        return self

    @classmethod
    def __new_deserialized__(cls, data, info):
        view_key, table_id, row_ids, cap, cap_next = data
        table_data, _ = info.shared_data(SHARED_DATA_KEY)
        if not info.remote_is_trusted:
            validate_cap(cap, view_key, table_id)
            table_data = None
        if not table_data:
            row_ids = cap_next = None
        # when we deserialize ourselves we may have more data than we need
        self = cls._create(view_key, table_id, row_ids, cap, cap_next, table_data)
        self._from_serialize = True
        return self

    def _fill_data(self):
        self._row_ids, self._cap_next, self._table_data = anvil.server.call(PREFIX + "next_page", self._cap)

    def _clear_cache(self):
        self._row_ids = self._table_data = self._cap_next = None

    # SERIALIZATION
    def _make_row_data(self, row_data, table_spec, compact=True):
        if type(row_data) is dict or compact:
            # this row didn't match our cache_spec so just send it
            # or we are list and we're compact because our cache_specs already match
            return row_data

        cache_spec = table_spec["cache"]
        # we are currently compact and we need to be a dict
        new_data = {CAP_KEY: row_data[-1]}
        iter_row_data = iter(row_data)

        new_data = {str(i): next(iter_row_data) for i, is_cached in enumerate(cache_spec) if is_cached}
        cap = next(iter_row_data)
        assert type(cap is Capability)
        new_data[CAP_KEY] = cap
        return new_data

    def _get_table_view_iter(self):
        if not self._from_serialize:
            # Fast Path - we were created from my_table.search() so the table_data is already minimal
            # i.e. we don't need to clean it based on table_specs
            return self._table_data.keys()

        # Slow Path - we're reserializing ourselves from a previous serialization
        # so we may have too much data if we were serialized with merged table_data
        table_view_keys = set()
        # walk the table_specs and insert the view_keys and table_ids we need
        _populate_table_views_ids(self._view_key, self._table_data, table_view_keys)
        return table_view_keys

    def _merge(self, g_table_data, local_data):
        if check_serialized(self, local_data):
            return

        table_view_keys = self._get_table_view_iter()

        for view_key in table_view_keys:
            g_view_data = init_view_data(view_key, g_table_data)
            l_view_data = self._table_data[view_key]

            l_table_spec, l_table_rows = l_view_data["spec"], l_view_data.get("rows", {})
            g_table_spec, g_table_rows = init_spec_rows(g_view_data, l_table_spec)

            g_cache_spec = g_table_spec["cache"]
            l_cache_spec = l_table_spec["cache"]
            cache_match = g_table_spec is l_table_spec or g_cache_spec == l_cache_spec

            for row_id, row_data in l_table_rows.items():
                if isinstance(row_data, Row):
                    # Ok we've already been created
                    # this is rare - we've consumed the search iterator and now we're serializing
                    # or we created this row from shared serialization data and we're now reserializing
                    row = row_data
                    row._merge_and_reduce(g_table_data, local_data)
                    continue

                g_row_data = g_table_rows.get(row_id, [])
                g_is_compact = cache_match and type(g_row_data) is list
                row_data = self._make_row_data(row_data, l_table_spec, compact=g_is_compact)
                merge_row_data(row_id, row_data, g_table_rows, g_table_spec, l_cache_spec)

    def __serialize__(self, info):
        table_data, local_data = info.shared_data(SHARED_DATA_KEY)
        row_ids = self._row_ids
        if table_data is None:
            row_ids = self._cap_next = None
        elif info.local_is_trusted and self._table_data is not None:
            self._merge(table_data, local_data)
        return [self._view_key, self._table_id, row_ids, self._cap, self._cap_next]

    def _make_partial_iterator(self, slice_=slice(None)):
        return PartialSearchIter(self, slice_)

    def __iter__(self):
        return self._make_partial_iterator()

    def __len__(self):
        if self._cap_next is None and self._row_ids is not None:
            return len(self._row_ids)
        return anvil.server.call(PREFIX + "get_length", self._cap)

    def __hash__(self):
        return hash((self._table_id, self._cap))

    def __eq__(self, other):
        if not isinstance(other, SearchIterator):
            return NotImplemented
        return self._cap == other._cap

    def __bool__(self):
        # because we have a __len__ and we can't suspend
        return True

    __nonzero__ = __bool__

    def to_csv(self):
        return anvil.server.call(PREFIX + "to_csv", self._cap)

    def delete_all_rows(self):
        result = anvil.server.call(PREFIX + "delete_all", self._cap)
        self._clear_cache()
        return result

    def __getitem__(self, idx):
        if self._row_ids is None:
            self._fill_data()

        if isinstance(idx, slice):
            slice_ = slice(as_slice_idx(idx.start), as_slice_idx(idx.stop), as_slice_idx(idx.step))
            return self._make_partial_iterator(slice_)
        else:
            slice_ = slice(as_idx(idx), None)
        try:
            return next(self._make_partial_iterator(slice_))
        except StopIteration:
            raise IndexError("search index out of range")


def as_idx(i, msg="search indices must be non-negative integers", can_be_none=False):
    if i is None and can_be_none:
        return None
    elif type(i) is int:
        pass
    elif hasattr(i, "__index__"):
        i = i.__index__()
    else:
        raise TypeError(msg)
    if i < 0:
        raise ValueError(msg)
    return i


def as_slice_idx(i):
    msg = "search slice indices must non-negative itegers (or None)"
    return as_idx(i, msg, True)


def _populate_table_views_ids(view_key, table_data, seen):
    # We might hold too much data if our table_data was from another serialization
    # If we're reserializing ourselves then this method prevents sending unnecessary data across the wire
    if view_key in seen:
        # prevent circular references
        return

    try:
        table_spec = table_data[view_key]["spec"]
    except KeyError:
        # Then these linked rows were not included in the data - probably uncached from the cache spec
        # don't try include this view_key when serializing the data
        return

    seen.add(view_key)
    cols = table_spec["cols"]

    for col in cols:
        view_key = col.get("view_key")
        if view_key is None:
            continue
        _populate_table_views_ids(view_key, table_data, seen)



import anvil.server
from anvil import *
from .exceptions import UserExists, AuthenticationFailed, EmailNotConfirmed, AccountIsNotEnabled, PasswordNotAcceptable, MFARequired, MFAException, PasswordResetRequested, TooManyPasswordFailures
from .config import get_client_config

def _to_user_row(user):
    if type(user) is list:
        from anvil.tables.v2._row import Row
        user = Row._create_from_trusted(*user)
    return user

def _to_row_ref(user):
    from anvil.tables import _get_config
    if _get_config().get("enable_v2"):
        from anvil.tables.v2._refs import to_ref
        user = to_ref(user)
    return user

#!suggestAttr(anvil.users,login_with_form)!0:

#!defFunction(anvil.users,_)!2: "Forget the current logged-in user" ["logout"]
def logout():
    anvil.server.call("anvil.private.users.logout")


anvil.server._register_exception_type("anvil.users.UserExists", UserExists)
anvil.server._register_exception_type("anvil.users.AuthenticationFailed", AuthenticationFailed)
anvil.server._register_exception_type("anvil.users.EmailNotConfirmed", EmailNotConfirmed)
anvil.server._register_exception_type("anvil.users.AccountIsNotEnabled", AccountIsNotEnabled)
anvil.server._register_exception_type("anvil.users.TooManyPasswordFailures", TooManyPasswordFailures)
anvil.server._register_exception_type("anvil.users.PasswordNotAcceptable", PasswordNotAcceptable)
anvil.server._register_exception_type("anvil.users.MFARequired", MFARequired)
anvil.server._register_exception_type("anvil.users.MFAException", MFAException)
anvil.server._register_exception_type("anvil.users.PasswordResetRequested", PasswordResetRequested)

#!defFunction(anvil.users,_,email,password,[remember=False])!2: "Log in with the specified email address and password. Raises anvil.users.AuthenticationFailed exception if the login failed.\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["login_with_email"]
def login_with_email(email, password, remember=False, mfa=None):
    u = anvil.server.call("anvil.private.users.login_with_email", email, password, remember=remember, mfa=mfa)
    return _to_user_row(u)

#!defFunction(anvil.users,_,email,password,[remember=False])!2: "Sign up for a new account with the specified email address and password. Raises anvil.users.UserExists if an account is already registered with this email address.\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["signup_with_email"]
def signup_with_email(email, password, remember=False):
    u = anvil.server.call("anvil.private.users.signup_with_email", email, password, remember=remember)
    return _to_user_row(u)

#!defFunction(anvil.users,_,email_address)!2: "Send a password-reset email to the specified user" ["send_password_reset_email"]
def send_password_reset_email(email):
    anvil.server.call("anvil.private.users.send_password_reset_email", email)

#!defFunction(anvil.users,_,email_address)!2: "Send a login link email to the specified user" ["send_token_login_email"]
def send_token_login_email(email):
    anvil.server.call("anvil.private.users.send_token_login_email", email)

#!defFunction(anvil.users,_,old_password,new_password)!2: "Reset the password for the current user" ["reset_password"]
def reset_password(old_password, new_password):
    anvil.server.call("anvil.private.users.reset_password", old_password, new_password)


if is_server_side():
    def get_user(allow_remembered=True):
        user = anvil.server.call("anvil.private.users.get_current_user", allow_remembered=allow_remembered)
        return _to_user_row(user)

    #!defFunction(anvil.users,_,user_row,[remember=False])!2: "Set the specified user object (a row from a Data Table) as the current logged-in user. It must be a row from the users table. By default, login status is not remembered between sessions." ["force_login"]
    def force_login(user, remember=False):
        u = anvil.server.call("anvil.private.users.force_login", _to_row_ref(user), remember=remember)
        return _to_user_row(u)

    def _fail(fname):
        def f(*args, **kwargs):
            raise Exception("You can't use " + fname + "() on the server (do it in form code instead)")
        return f

    for n in ["login_with_google", "signup_with_google", "login_with_facebook", "signup_with_facebook", "login_with_microsoft", "signup_with_microsoft", "login_with_saml", "signup_with_saml", "login_with_raven", "signup_with_raven", "login_with_form", "signup_with_form"]:
        globals()[n] = _fail(n)

else:
    from . import mfa

    #!defFunction(anvil.users,_,[allow_remembered=True])!2: "Get the row from the users table that corresponds to the currently logged-in user. If allow_remembered is true (the default), the user may have logged in in a previous session. Returns None if no user is logged in." ["get_user"]
    def get_user(allow_remembered=True):
        try:
            user = anvil.server.call("anvil.private.users.get_current_user", allow_remembered=allow_remembered)
        except MFARequired:
            # This should only happen if the user has arrived with a login token and needs to configure MFA
            mfa.configure_mfa_with_form()
            return None
        except PasswordResetRequested:
            change_password_with_form(False)
            user = anvil.server.call("anvil.private.users.get_current_user", allow_remembered=allow_remembered)
        
        return _to_user_row(user)


    def force_login(user, remember=False):
        raise Exception("You can only use force_login() in server modules")

    #!defFunction(anvil.users,!,[additional_scopes],[remember=False])!2: "Log in with a Google account. Prompts the user to authenticate with Google, then logs in with their Google email address (if that user exists). Returns None if the login was cancelled or we have no record of this user.\n\nadditional_scopes: If supplied, these are passed on to anvil.google.auth.login().\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["login_with_google"]
    def login_with_google(additional_scopes=None, remember=False):
        if not get_client_config().get("use_google", False):
            raise Exception("Google login is not enabled")

        import anvil.google.auth
        if anvil.google.auth.login(additional_scopes):
            u = anvil.server.call("anvil.private.users.login_with_google", remember=remember)
            return _to_user_row(u)

    #!defFunction(anvil.users,!,[additional_scopes],[remember=False])!2: "Sign up for a new account with the email address associated with the user's Google account. Prompts the user to authenticate with Google, then registers a new user with that email address. Raises anvil.users.UserExists if this email address is already registered; returns new user or None if cancelled.\n\nadditional_scopes: If supplied, these are passed on to anvil.google.auth.login().\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["signup_with_google"]
    def signup_with_google(additional_scopes=None, remember=False):
        if not get_client_config().get("use_google", False):
            raise Exception("Google signup is not enabled")

        if not get_client_config().get("allow_signup"):
            raise Exception("New user signup is not enabled")

        import anvil.google.auth
        if anvil.google.auth.login(additional_scopes):
            u = anvil.server.call("anvil.private.users.signup_with_google", remember=remember)
            return _to_user_row(u)

    #!defFunction(anvil.users,!,[additional_scopes],[remember=False])!2: "Log in with a Facebook account. Prompts the user to authenticate with Facebook, then logs in with their Facebook email address (if that user exists). Returns None if the login was cancelled or we have no record of this user.\n\nadditional_scopes: If supplied, these are passed on to anvil.facebook.auth.login().\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["login_with_facebook"]
    def login_with_facebook(additional_scopes=None, remember=False):
        if not get_client_config().get("use_facebook"):
            raise Exception("Facebook login is not enabled")

        import anvil.facebook.auth
        if anvil.facebook.auth.login(additional_scopes):
            u = anvil.server.call("anvil.private.users.login_with_facebook", remember=remember)
            return _to_user_row(u)

    #!defFunction(anvil.users,!,[additional_scopes],[remember=False])!2: "Sign up for a new account with the email address associated with the user's Facebook account. Prompts the user to authenticate with Facebook, then registers a new user with that email address. Raises anvil.users.UserExists if this email address is already registered; returns new user or None if cancelled.\n\nadditional_scopes: If supplied, these are passed on to anvil.facebook.auth.login().\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["signup_with_facebook"]
    def signup_with_facebook(additional_scopes=None, remember=False):
        if not get_client_config().get("use_facebook"):
            raise Exception("Facebook signup is not enabled")

        if not get_client_config().get("allow_signup"):
            raise Exception("New user signup is not enabled")

        import anvil.facebook.auth
        if anvil.facbeook.auth.login(additional_scopes):
            u = anvil.server.call("anvil.private.users.signup_with_facebook", remember=remember)
            return _to_user_row(u)

    #!defFunction(anvil.users,!,[additional_scopes],[remember=False])!2: "Log in with a Microsoft account. Prompts the user to authenticate with Microsoft, then logs in with their Microsoft email address (if that user exists). Returns None if the login was cancelled or we have no record of this user.\n\nadditional_scopes: If supplied, these are passed on to anvil.microsoft.auth.login().\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["login_with_microsoft"]
    def login_with_microsoft(additional_scopes=None, remember=False):
        if not get_client_config().get("use_microsoft"):
            raise Exception("Microsoft login is not enabled")

        import anvil.microsoft.auth
        if anvil.microsoft.auth.login(additional_scopes):
            u = anvil.server.call("anvil.private.users.login_with_microsoft", remember=remember)
            return _to_user_row(u)

    #!defFunction(anvil.users,!,[additional_scopes],[remember=False])!2: "Sign up for a new account with the email address associated with the user's Microsoft account. Prompts the user to authenticate with Microsoft, then registers a new user with that email address. Raises anvil.users.UserExists if this email address is already registered; returns new user or None if cancelled.\n\nadditional_scopes: If supplied, these are passed on to anvil.microsoft.auth.login().\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["signup_with_microsoft"]
    def signup_with_microsoft(additional_scopes=None, remember=False):
        if not get_client_config().get("use_microsoft"):
            raise Exception("Microsoft signup is not enabled")

        if not get_client_config().get("allow_signup"):
            raise Exception("New user signup is not enabled")

        import anvil.microsoft.auth
        if anvil.microsoft.auth.login(additional_scopes):
            u = anvil.server.call("anvil.private.users.signup_with_microsoft", remember=remember)
            return _to_user_row(u)

    #!defFunction(anvil.users,!,[remember=False])!2: "Log in via a SAML Identity Provider. Prompts the user to authenticate with SAML, then logs in with their email address (if that user exists). Returns None if the login was cancelled or we have no record of this user.\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["login_with_saml"]
    def login_with_saml(remember=False):
        if not get_client_config().get("use_saml"):
            raise Exception("SAML login is not enabled")

        import anvil.saml.auth
        if anvil.saml.auth.login():
            u = anvil.server.call("anvil.private.users.login_with_saml", remember=remember)
            return _to_user_row(u)

    #!defFunction(anvil.users,!,[remember=False])!2: "Sign up for a new account with the email address associated with the user's SAML account. Prompts the user to authenticate via SAML, then registers a new user with that email address. Raises anvil.users.UserExists if this email address is already registered; returns new user or None if cancelled.\n\nBy default, login status is not remembered between sessions; set remember=True to remember login status." ["signup_with_saml"]
    def signup_with_saml(additional_scopes=None, remember=False):
        if not get_client_config().get("use_saml"):
            raise Exception("SAML signup is not enabled")

        if not get_client_config().get("allow_signup"):
            raise Exception("New user signup is not enabled")

        import anvil.saml.auth
        if anvil.saml.auth.login(additional_scopes):
            u = anvil.server.call("anvil.private.users.signup_with_saml", remember=remember)
            return _to_user_row(u)

    # Disable documentation for raven functions for now.
    #defFunction(anvil.users,!,[remember=False])!2: "Log in with a Raven account. Prompts the user to authenticate with Raven, then logs in with their Raven account (if that user exists). Returns None if the login was cancelled or we have no record of this user. By default, login status is not remembered between sessions." ["login_with_raven"]
    def login_with_raven(remember=False):
        if not get_client_config().get("use_raven", False):
            raise Exception("Raven login is not enabled")

        import raven.auth
        if raven.auth.login():
            u = anvil.server.call("anvil.private.users.login_with_raven", remember=remember)
            return _to_user_row(u)

    #defFunction(anvil.users,!,[remember=False])!2: "Sign up for a new account with the email address associated with the user's Raven account. Prompts the user to authenticate with Raven, then registers a new user with that email address. Raises anvil.users.UserExists if this email address is already registered; returns new user or None if cancelled. By default, login status is not remembered between sessions." ["signup_with_raven"]
    def signup_with_raven(remember=False):
        if not get_client_config().get("use_raven", False):
            raise Exception("Raven signup is not enabled")

        if not get_client_config().get("allow_signup"):
            raise Exception("New user signup is not enabled")

        import raven.auth
        if raven.auth.login():
            u = anvil.server.call("anvil.private.users.signup_with_raven", remember=remember)
            return _to_user_row(u)


    _label_style = {}

    def _email_token_login_with_form(initial_email=""):
        panel = LinearPanel()

        email_box = TextBox(placeholder="Email address", text=initial_email)
        panel.add_component(Label(text="Enter your registered email address, and we will send you a \"magic link\" by email that will log you in."))
        panel.add_component(email_box)

        def show(**e):
            email_box.focus()
            email_box.select()

        email_box.set_event_handler("show", show)
        email_box.set_event_handler("pressed_enter", lambda **e: panel.raise_event("x-close-alert", value=True))

        if alert(panel, title="Login by email", buttons=[("Send email", True, 'success'),("Cancel", False, "default")], dismissible=True):
            return email_box.text




    #!defFunction(anvil.users,!,[remember_by_default=True],[allow_cancel=False])!2: "Display a sign-up form allowing a user to create a new account. Returns the new user object, or None if cancelled.\n\nremember_by_default: if True, the 'remember me' checkbox will be enabled by default.\n\nallow_cancel: if True, the signup form has a Cancel button that the user can use to dismiss the form." ["signup_with_form"]
    def signup_with_form(_link_back_to_login_on_already_exists=False,remember_by_default=True, allow_cancel=False, initial_email="", initial_password=""):
        if not get_client_config().get("allow_signup"):
            raise Exception("New user signup is not enabled")

        lp = LinearPanel()
        email_box = None
        passwd_box = None
        remember_me_checkbox = None

        def email_pressed_enter(**kws):
            if passwd_box and len(passwd_box) > 0:
                passwd_box[0].focus()

        def passwd_1_pressed_enter(**kws):
            if passwd_box and len(passwd_box) > 1:
                passwd_box[1].focus()

        def passwd_2_pressed_enter(**kws):
            lp.raise_event('x-close-alert', value='sign-up')

        some_method_available = False

        if get_client_config().get("use_email", False):
            some_method_available = True
            lp.add_component(Label(text="Email:", **_label_style))
            email_box = TextBox(placeholder="address@example.com", text=initial_email)
            email_box.set_event_handler("pressed_enter", email_pressed_enter)
            lp.add_component(email_box)

        if get_client_config().get("use_email", False):
            some_method_available = True
            passwd_box = [TextBox(hide_text=True, placeholder=p) for p in ["password", "repeat password"]]
            lp.add_component(Label(text="Password:", **_label_style))
            passwd_box[0].set_event_handler("pressed_enter", passwd_1_pressed_enter)
            passwd_box[0].text = initial_password
            lp.add_component(passwd_box[0])
            lp.add_component(Label(text="Retype password:", **_label_style))
            passwd_box[1].set_event_handler("pressed_enter", passwd_2_pressed_enter)
            lp.add_component(passwd_box[1])

        if get_client_config().get("use_google", False):
            some_method_available = True
            def google_login(**evt):
                import anvil.google.auth
                if anvil.google.auth.login():
                    lp.raise_event('x-close-alert', value='google')
            lnk = Link(spacing_above="large", spacing_below="none")
            lnk.add_component(Image(source=anvil._get_anvil_cdn_origin() + "/runtime/img/google-signin-buttons/btn_google_signin_light_normal_web.png", display_mode="original_size"))
            lnk.set_event_handler("click", google_login)
            lp.add_component(lnk)

        if get_client_config().get("use_facebook"):
            some_method_available = True
            def facebook_login(**evt):
                import anvil.facebook.auth
                if anvil.facebook.auth.login():
                    lp.raise_event('x-close-alert', value='facebook')
            b = Button(text="Sign up with Facebook", icon="fa:facebook", icon_align="left")
            lp.add_component(b)
            b.set_event_handler("click", facebook_login)

        if get_client_config().get("use_microsoft"):
            some_method_available = True
            def microsoft_login(**evt):
                import anvil.microsoft.auth
                if anvil.microsoft.auth.login():
                    lp.raise_event('x-close-alert', value='microsoft')
            b = Button(text="Sign up with Microsoft", icon="fa:windows", icon_align="left")
            lp.add_component(b)
            b.set_event_handler("click", microsoft_login)

        if get_client_config().get("use_saml"):
            some_method_available = True
            def saml_login(**evt):
                import anvil.saml.auth
                if anvil.saml.auth.login():
                    lp.raise_event('x-close-alert', value='saml')
            b = Button(text="Sign up via SAML", icon="fa:lock", icon_align="left")
            lp.add_component(b)
            b.set_event_handler("click", saml_login)

        if get_client_config().get("use_raven", False):
            some_method_available = True
            import raven.auth
            def raven_login(**evt):
                if raven.auth.login():
                    lp.raise_event('x-close-alert', value='raven')
            b = Button(text="Sign up with Raven", icon="fa:lock", icon_align="left")
            lp.add_component(b)
            b.set_event_handler("click", raven_login)

        if not some_method_available:
            raise Exception("This app has no supported sign-in methods. Please check settings in the Users Service configuration.")

        error_lbl = Label(foreground="red", bold=True, spacing_below="none")
        lp.add_component(error_lbl)

        log_in_instead_link = Link(text="Log in instead", visible=False,
                                   icon="fa:chevron-right", icon_align="right", spacing_above="none")
        def log_in_instead(**evt):
            lp.raise_event('x-close-alert', value=None)
        log_in_instead_link.set_event_handler("click", log_in_instead)
        if _link_back_to_login_on_already_exists:
            lp.add_component(log_in_instead_link)


        if get_client_config().get("allow_remember_me", False) and not get_client_config().get("confirm_email", False):
            remember_me_checkbox = CheckBox(text="Remember me", checked=remember_by_default)
            lp.add_component(remember_me_checkbox)

        if email_box and passwd_box:
            lp.set_event_handler("show", lambda **e: (email_box.focus() if not initial_email else (passwd_box[0].focus() if not passwd_box[0].text else passwd_box[1].focus())))

        attempts = 0
        while True:
            if passwd_box and attempts > 0:
                for pb in passwd_box:
                    pb.text = ""

            attempts += 1

            maybe_cancel_button = [("Cancel", None)] if allow_cancel else []

            if get_client_config().get("use_email", False):
                ar = alert(lp, title="Sign Up", buttons=[("Sign Up", 'sign-up', "primary")] + maybe_cancel_button, dismissible=allow_cancel)
            else:
                ar = alert(lp, title="Sign Up", buttons=maybe_cancel_button, dismissible=allow_cancel)

            if not ar:
                return None

            # TODO require certain fields and include them in the sign-up call
 
            remember = (remember_me_checkbox and remember_me_checkbox.checked)

            try:
                if ar == 'google':
                    user = anvil.server.call("anvil.private.users.signup_with_google", remember=remember)
                elif ar == 'facebook':
                    user = anvil.server.call("anvil.private.users.signup_with_facebook", remember=remember)
                elif ar == 'microsoft':
                    user = anvil.server.call("anvil.private.users.signup_with_microsoft", remember=remember)
                elif ar == 'saml':
                    user = anvil.server.call("anvil.private.users.signup_with_saml", remember=remember)
                elif ar == 'raven':
                    user = anvil.server.call("anvil.private.users.signup_with_raven", remember=remember)
                elif ar == 'sign-up' and passwd_box:
                    if len(email_box.text) < 5 or "@" not in email_box.text or "." not in email_box.text:
                        error_lbl.text = "Enter an email address"
                        continue
                    if passwd_box[1].text != passwd_box[0].text:
                        error_lbl.text = "Passwords do not match"
                        continue

                    try:
                        user = anvil.server.call("anvil.private.users.signup_with_email", email_box.text, passwd_box[0].text, remember=remember)
                    except MFARequired:
                        mfa_method, _ = mfa._configure_mfa(email_box.text, None, False, [("Cancel", None)], "Sign up")
                        if mfa_method:
                            user = anvil.server.call("anvil.private.users.signup_with_email", email_box.text, passwd_box[0].text, mfa_method=mfa_method, remember=remember)
                        else:
                            continue

                    if get_client_config().get("confirm_email", False):
                        alert("We've sent a confirmation email to " + email_box.text + ". Open your inbox and click the link to complete your signup.", title="Confirm your Email", buttons=[("OK", None, "primary")])
                    
                    return _to_user_row(user)

                else:
                    raise Exception("Invalid configuration for Users service")

            except UserExists as e:
                error_lbl.text = str(e.args[0])
                log_in_instead_link.visible = True
                continue

            except PasswordNotAcceptable as e:
                error_lbl.text = str(e.args[0])
                log_in_instead_link.visible = False
                continue

            return user

    #!defFunction(anvil.users,!,[show_signup_option=True],[remember_by_default=True],[allow_remembered=True],[allow_cancel=False])!2: "Display a login form and allow user to log in. Returns user object if logged in, or None if cancelled.\n\nshow_signup_option: if True, the form will also show the option to sign up for a new account.\n\nremember_by_default: if True, the 'remember me' checkbox will be enabled by default.\n\nallow_remembered: if False, users with remembered login status will still be required to log in.\n\nallow_cancel: if True, the login form has a Cancel button that the user can use to dismiss the form." ["login_with_form"]
    def login_with_form(show_signup_option=True,remember_by_default=True, allow_remembered=True, allow_cancel=False):
        
        if allow_remembered:
           u = get_user()
           if u:
               return u

        lp = LinearPanel()
        email_box = None
        passwd_box = None
        remember_me_checkbox = None

        def focus_email(**kws):
            if email_box:
                email_box.focus()

        def focus_password(**kws):
            if passwd_box:
                passwd_box.focus()

        def close_alert(**kws):
            lp.raise_event('x-close-alert', value='login')

        some_method_available = False
        if get_client_config().get("use_email", False):
            some_method_available = True

            last_email = anvil.server.call("anvil.private.users.get_last_login_email")

            email_box = TextBox(placeholder="email@address.com", text=last_email)
            passwd_box = TextBox(placeholder="password", hide_text=True, spacing_below="none")

            email_box.set_event_handler("pressed_enter", focus_password)
            passwd_box.set_event_handler("pressed_enter", close_alert)

            if last_email is None:
                lp.set_event_handler("show", focus_email)
            else:
                lp.set_event_handler("show", focus_password)

            lp.add_component(Label(text="Email:", **_label_style))
            lp.add_component(email_box)
            lp.add_component(Label(text="Password:", **_label_style))
            lp.add_component(passwd_box)
            reset_link = Link(text="Forgot your password?", font_size=12, spacing_above="none", align="right")
            reset_link.set_event_handler('click', lambda **e: lp.raise_event('x-close-alert', value='reset_password'))
            lp.add_component(reset_link)

        if get_client_config().get("use_google", False):
            some_method_available = True
            def google_login(**evt):
                import anvil.google.auth
                if anvil.google.auth.login():
                    lp.raise_event('x-close-alert', value='google')
                
            lnk = Link(spacing_above="large", spacing_below="none")
            lnk.add_component(Image(source=anvil._get_anvil_cdn_origin() + "/runtime/img/google-signin-buttons/btn_google_signin_light_normal_web.png", display_mode="original_size"))
            lnk.set_event_handler("click", google_login)
            lp.add_component(lnk)

        if get_client_config().get("use_facebook"):
            some_method_available = True
            def facebook_login(**evt):
                import anvil.facebook.auth
                if anvil.facebook.auth.login():
                    lp.raise_event('x-close-alert', value='facebook')
                
            b = Button(text="Log in with Facebook", icon="fa:facebook", icon_align="left")
            b.set_event_handler('click', facebook_login)
            lp.add_component(b)

        if get_client_config().get("use_microsoft"):
            some_method_available = True
            def microsoft_login(**evt):
                import anvil.microsoft.auth
                if anvil.microsoft.auth.login():
                    lp.raise_event('x-close-alert', value='microsoft')
                
            b = Button(text="Log in with Microsoft", icon="fa:windows", icon_align="left")
            b.set_event_handler('click', microsoft_login)
            lp.add_component(b)

        if get_client_config().get("use_saml"):
            some_method_available = True
            def saml_login(**evt):
                import anvil.saml.auth
                if anvil.saml.auth.login():
                    lp.raise_event('x-close-alert', value='saml')
                
            b = Button(text="Log in via SAML", icon="fa:lock", icon_align="left")
            b.set_event_handler('click', saml_login)
            lp.add_component(b)

        if get_client_config().get("use_raven", False):
            some_method_available = True
            def raven_login(**evt):
                import raven.auth
                if raven.auth.login():
                    lp.raise_event('x-close-alert', value='raven')
                
            b = Button(text="Log in with Raven", icon="fa:lock", icon_align="left")
            b.set_event_handler('click', raven_login)
            lp.add_component(b)

        if get_client_config().get("use_token", False):
            some_method_available = True
            b = Link(text="Send a login link by email", icon="fa:envelope", icon_align="left", align="center")
            b.set_event_handler('click', lambda **e: lp.raise_event('x-close-alert', value='email_token'))
            lp.add_component(b)

        if not some_method_available:
            raise Exception("This app has no supported sign-in methods. Please check settings in the Users Service configuration.")

        error_lbl = Label(foreground="red", bold=True)
        lp.add_component(error_lbl)

        if get_client_config().get("allow_signup") and show_signup_option:
            def open_signup(**evt):
                lp.raise_event('x-close-alert', value='sign-up')
            signup_link = Link(text="Sign up for a new account", icon="fa:user-plus")
            signup_link.set_event_handler('click', open_signup)
            lp.add_component(signup_link)

        if get_client_config().get("allow_remember_me", False):
            remember_me_checkbox = CheckBox(text="Remember me", checked=remember_by_default)
            lp.add_component(remember_me_checkbox)

        while True:
            if passwd_box:
                passwd_box.text = ""

            maybe_cancel_button = [("Cancel", None)] if allow_cancel else []

            if get_client_config().get("use_email", False):
                ar = alert(lp, title="Log In", buttons=[("Log In", 'login', 'success')] + maybe_cancel_button, dismissible=allow_cancel)
            else:
                ar = alert(lp, title="Log In", buttons=maybe_cancel_button, dismissible=allow_cancel)

            remember = (remember_me_checkbox and remember_me_checkbox.checked)
            try:
                if ar == 'google':
                    user = anvil.server.call("anvil.private.users.login_with_google", remember=remember)
                    if user or allow_cancel:
                        return user
                elif ar == 'facebook':
                    user = anvil.server.call("anvil.private.users.login_with_facebook", remember=remember)
                    if user or allow_cancel:
                        return user
                elif ar == 'microsoft':
                    user = anvil.server.call("anvil.private.users.login_with_microsoft", remember=remember)
                    if user or allow_cancel:
                        return user
                elif ar == 'saml':
                    user = anvil.server.call("anvil.private.users.login_with_saml", remember=remember)
                    if user or allow_cancel:
                        return user
                elif ar == 'raven':
                    user = anvil.server.call("anvil.private.users.login_with_raven", remember=remember)
                    if user or allow_cancel:
                        return user
                elif ar == 'email_token':
                    target_email = _email_token_login_with_form(email_box.text if email_box else "")
                    if target_email:
                        send_token_login_email(target_email)
                        alert("An email with a login link has been sent to you. You can now close this window.", buttons=[], dismissible=False)                        
                elif ar == 'reset_password':
                    reset_email_box = TextBox(placeholder="email@address.com", text=email_box.text)
                    pnl = LinearPanel()
                    pnl.add_component(reset_email_box)
                    if alert(pnl, title="Reset password by email", buttons=[("OK", True, "primary"), ("Cancel", False, "default")]):
                        send_password_reset_email(reset_email_box.text)
                        error_lbl.text = "Requested password reset for " + reset_email_box.text + ". Check your email."
                    else:
                        error_lbl.text = ""
                    # Continue around loop
                elif ar == 'login':

                    try:
                        return login_with_email(email_box.text, passwd_box.text, remember=remember)
                    except MFARequired:
                        
                        r = mfa.mfa_login_with_form(email_box.text, passwd_box.text)
                        
                        if r == 'reset_mfa':
                            mfa.send_mfa_reset_email(email_box.text)
                            error_lbl.text = "Requested 2-factor authentication reset for " + email_box.text + ". Check your email."
                        elif r == None:
                            if allow_cancel:
                                return None
                            # Else continue around the loop
                        else:
                            # We got an MFA dict. Log in with it.
                            return login_with_email(email_box.text, passwd_box.text, mfa=r, remember=remember)

                elif ar == 'sign-up':
                    if signup_with_form(_link_back_to_login_on_already_exists=True, allow_cancel=True, initial_email=email_box.text if email_box else "", initial_password=passwd_box.text if passwd_box else ""):
                        user = get_user(allow_remembered=False)
                        if user:
                            return user
                    # else continue around the loop
                else:
                    return None
            except AuthenticationFailed as e:
                error_lbl.text = e.args[0]

    #!defFunction(anvil.users,_,[require_old_password=True])!2: "Display a form allowing the current user to reset their password. " ["change_password_with_form"]
    def change_password_with_form(require_old_password=True):

        err = None
        while True:
            panel = LinearPanel()

            old_pwd = TextBox(hide_text=True, placeholder="Old password")
            new_pwd = [TextBox(hide_text=True, placeholder="New password"), TextBox(hide_text=True, placeholder="Confirm new password")]

            if require_old_password:
                panel.add_component(old_pwd)
                old_pwd.set_event_handler("show", lambda **e: old_pwd.focus())
                old_pwd.set_event_handler("pressed_enter", lambda **e: new_pwd[0].focus())
            else:
                new_pwd[0].set_event_handler("show", lambda **e: new_pwd[0].focus())

            new_pwd[0].set_event_handler("pressed_enter", lambda **e: new_pwd[1].focus())
            new_pwd[1].set_event_handler("pressed_enter", lambda **e: panel.raise_event("x-close-alert", value=True))

            panel.add_component(new_pwd[0])
            panel.add_component(new_pwd[1])

            if err:
                panel.add_component(Label(text=err, foreground="red"))

            r = alert(panel, title="Change password" if require_old_password else "Reset password", buttons=[("Change" if require_old_password else "Reset", True, "success"), ("Cancel", False, "default")], dismissible=True)

            if r:
                if new_pwd[0].text == new_pwd[1].text:
                    try:
                        # Reset password for the currently logged-in user
                        reset_password(old_pwd.text, new_pwd[0].text)
                        return
                    except AuthenticationFailed as e:
                        err = str(e.args[0])
                    except PasswordNotAcceptable as e:
                        err = str(e.args[0])
                else:
                    err = "Passwords do not match"
            else:
                anvil.server.call("anvil.private.users.cancel_password_reset")
                return


    #!defFunction(anvil.users,_)!2: "Display a form allowing the current user to configure their account. The form contains links for password reset and two-factor authentication configuration." ["configure_account_with_form"]
    def configure_account_with_form():

        if not get_user():
            raise Exception("Cannot configure user account: Not logged in.")

        while True:
            panel = LinearPanel()

            reset_password_link = Link(text="Change password")
            configure_mfa_link = Link(text="Configure two-factor authentication")

            reset_password_link.set_event_handler("click", lambda **e: panel.raise_event("x-close-alert", value="reset_password"))
            configure_mfa_link.set_event_handler("click", lambda **e: panel.raise_event("x-close-alert", value="configure_mfa"))

            panel.add_component(reset_password_link)
            panel.add_component(configure_mfa_link)

            choice = alert(panel, title="Configure Account", buttons=[("Done", None, "success")], dismissible=True)

            if choice == "reset_password":
                change_password_with_form()
            elif choice == "configure_mfa":
                mfa.configure_mfa_with_form(True)
            else:
                break
